/*=======================================================================*/
/*  q_arith                                                              */
/*     FITSǡλѱ黻򤹤                                        */
/*=======================================================================*/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<ctype.h>
#include<math.h>
#include<q_lib.h>
/*=======================================================================*/
void print_usage(char *);
int check_command_line(int,char **,char **,char **,char **,char *);
int is_number(char *s);
int calc_frame_num(FITS_DATA *,char ,char *,FITS_DATA *);
int calc_frame_frame(FITS_DATA *,char ,FITS_DATA *,FITS_DATA *);
/*=======================================================================*/
void print_usage(char *s)
{
char *ss;

  ss=strip_name(s);
  fprintf(stdout,
"NAME\n"
"       %s - Arithmatic operation of FITS Image\n\n"
"SYNOPSIS\n"
"       q_arith file_1 operator file_2 out_file_name\n"
"       q_arith file_1 operator number out_file_name\n\n"
"DESCRIPTION\n"
"       q_arith does arithmatic operation of FITS Image.\n\n"
"       The first style does arithmatic operation using two input images\n"
"       and output an FITS image. Two inputs should have the same sizes of\n"
"       X, Y and W axis. Size of Z axis shuld be the same or at least one\n"
"       input file has size = 1. In the later case, operation for different\n"
"       Z axis positions and uniqe Z axis positions produce output with \n"
"       the same size of Z axis of larger input image.\n\n"
"       The second style does arithmatic operation with image and number.\n\n"
"       operator is either - + * / x s. * and x is the same. s is defined as\n"
"       A s B = sqrt(A*A+B*B).\n\n"
"       Output file has the same size with input file, and data type is double.\n\n"
,ss);
}
/*=======================================================================*/
int main(int argc, char ** argv)
{
char *file_1,*file_2,*out_file_name;
char  operator ;
FITS_DATA fits1,fits2,fits3 ;
int status=0;
int mode;
int z,w;
LONGLONG xy; 

/* Check command line */
  if(check_command_line(argc,argv,&file_1,&file_2,&out_file_name,&operator)!=0){
    fprintf(stdout,"Error @ check_command_line\n");exit(1);}

/* Open input file and out put files */
  if(openfits(&fits1,file_1)!=0){ exit(1); }
  if(is_number(file_2)){
    mode=0;
  }else{
    mode=1;
    if(openfits(&fits2,file_2)!=0){ exit(1); }
  }
  if(createfits(&fits3,out_file_name)!=0){ exit(1); }

/* Check the input data parameters */
  if(prep_readfits(&fits1)!=0){ exit(1); }
  if(mode==1){ 
    if(prep_readfits(&fits2)!=0){ exit(1); } 
  }

/* Check data type and make new data type*/
  if(mode==1){
  /* X dim must be same */
    if(fits1.naxes[0]!=fits2.naxes[0]){
      fprintf(stdout,"NAXIS1 differ\n");exit(1);
    }
  /* Y dim must be same */
    if(fits1.naxes[1]!=fits2.naxes[1]){
      fprintf(stdout,"NAXIS2 differ\n");exit(1);
    }
  /* Z dim  same or either 1 */
    if(fits1.naxes[2]==fits2.naxes[2]) { fits3.naxes[2]=fits1.naxes[2];}
    else if(fits1.naxes[2]>fits2.naxes[2] && fits2.naxes[2]==1 ) 
                                       { fits3.naxes[2]=fits1.naxes[2];}
    else if(fits2.naxes[2]>fits1.naxes[2] && fits1.naxes[2]==1 ) 
                                       { fits3.naxes[2]=fits2.naxes[2];}
    else { fprintf(stdout,"NAXIS3 differ\n");exit(1);}
  /* W dim must be same */
    if(fits1.naxes[3]!=fits2.naxes[3]){
      fprintf(stdout,"NAXIS4 differ\n");exit(1);
    }
  /* NAX */
    if(fits1.nax>fits2.nax) fits3.nax=fits1.nax ;
    else fits3.nax=fits2.nax ;
  }else{
    fits3.naxes[2]=fits1.naxes[2];
    fits3.nax=fits1.nax ;
  }
  fits3.naxes[0]=fits1.naxes[0];
  fits3.naxes[1]=fits1.naxes[1];
  fits3.naxes[3]=fits1.naxes[3];
 /* bp */
  fits3.bp = DOUBLE_IMG ;

/* Allocate memory for data */
  if(alloc_memory_for_data(&fits3)!=0) exit(1);
  xy = ((LONGLONG)fits3.naxes[0])*((LONGLONG)fits3.naxes[1]);
/* Write header  إå¤˻Ĥƽ񤭼̤ */
  if(convert_all_head(&fits1,&fits3)!=0){
    fprintf(stdout,"Error @ convert_all_head");
    exit(1);
  }

/* Calcuration */
  for(w=1;w<=fits3.naxes[3];++w){
  for(z=1;z<=fits3.naxes[2];++z){
    if(z==1 || fits1.naxes[2]==fits3.naxes[2]){
      if(read_2d_data(&fits1,z,w)!=0) { exit(1); }
    }
    if(mode==1){
      if(z==1 || fits2.naxes[2]==fits3.naxes[2]){
        if(read_2d_data(&fits2,z,w)!=0) { exit(1); }
      }
    }
    if(mode==0){
      if(calc_frame_num(&fits1,operator,file_2,&fits3)!=0){exit(1);}
    }else{
      if(calc_frame_frame(&fits1,operator,&fits2,&fits3)!=0){exit(1);}
    }
    if(write_2d_data(&fits3,z,w)!=0){exit(1);}
  }}

/* By */
  fits_close_file(fits3.fp,&status);

  exit(0);
}
int calc_frame_num(FITS_DATA *d1,char op,char *num,FITS_DATA *d3)
{
int x,y,nx,ny;
double w1,w2;

  if(d3->bp != DOUBLE_IMG ){
    fprintf(stdout,"Error @ calc_frame_num bp\n") ; return(1);
  }
  nx = d3->naxes[0]; ny = d3->naxes[1];
  sscanf(num,"%lf",&w2) ;
  for(y=0;y<ny;++y){
  for(x=0;x<nx;++x){
    w1 = d_data(d1,x,y);
    if(op == '+')      {((double *)d3->data)[x+y*nx] = w1 + w2 ;
    }else if(op == '-'){((double *)d3->data)[x+y*nx] = w1 - w2 ;
    }else if(op == '*'){((double *)d3->data)[x+y*nx] = w1 * w2 ;
    }else if(op == '/'){
      if(w2!=0.) ((double *)d3->data)[x+y*nx] = w1 / w2 ;
      else       ((double *)d3->data)[x+y*nx] = HUGE_VAL ;
    }else if(op == 's'){((double *)d3->data)[x+y*nx] = sqrt(w1*w1+w2*w2);
    }else{fprintf(stdout,"Op Error at calc_frame_num op\n");return(1);}
  }}
  return(0);
}
int calc_frame_frame(FITS_DATA *d1,char op,FITS_DATA *d2,FITS_DATA *d3)
{
int x,y,nx,ny;
double w1,w2;

  if(d3->bp != DOUBLE_IMG ){
    fprintf(stdout,"Error @ calc_frame_num bp\n") ; return(1);
  }
  nx = d3->naxes[0]; ny = d3->naxes[1];
  for(y=0;y<ny;++y){
  for(x=0;x<nx;++x){
    w1 = d_data(d1,x,y);
    w2 = d_data(d2,x,y);
    if(op == '+')      {((double *)d3->data)[x+y*nx] = w1 + w2 ;
    }else if(op == '-'){((double *)d3->data)[x+y*nx] = w1 - w2 ;
    }else if(op == '*'){((double *)d3->data)[x+y*nx] = w1 * w2 ;
    }else if(op == '/'){
      if(w2!=0.) ((double *)d3->data)[x+y*nx] = w1 / w2 ;
      else       ((double *)d3->data)[x+y*nx] = HUGE_VAL ;
    }else if(op == 's'){((double *)d3->data)[x+y*nx] = sqrt(w1*w1+w2*w2);
    }else{fprintf(stdout,"Op Error at calc_frame_num op\n");return(1);}
  }}
  return(0);
}
/*=======================================================================*/
int check_command_line(int argc,char ** argv ,char ** fname_1,char ** fname_2,
                       char ** fname_3,char *operator)
{
  if(argc!=5 ){
    print_usage(argv[0]); return(1);}

  *fname_1 = argv[1] ;
  *fname_2 = argv[3] ;
  *fname_3 = argv[4] ;
  sscanf(argv[2],"%c",operator);
  if(*operator == 'x') *operator = '*' ;
  if(!(
    (*operator=='+') || (*operator=='-') || (*operator=='*') ||
    (*operator=='/') || (*operator=='s')
    )){fprintf(stdout,"operation error\n");return(-1);}

  set_cmd_hist(argc,argv);

  return(0);
}
/*=======================================================================*/
/* ʸ󤬿ϤɤȽ  
     ͤλ򤷤ƤʸǤ1֤Ǥʤ0֤
     ͤλȤ
        ǽʸ '+' or '-' or '.' or 0-9
        ʤȤ1ʸ 0-9
        0-9  '+' '-' '.' 'e' ʳʸϤäƤϤʤʤ
        '.''e'ϺǤ1ʸ
        '+''-'Ƭ'e'ľˤäƤ褤
        'e'θˤ'.'äƤϥ
        'e'θˤϾʤȤ1ʸ 0-9 
        'e'θ˵0-93ʸޤ
     Ȥ§                                           */
int is_number(char *s)
{
char *c;
int i,k,e;
/* ǽʸ */
  if(*s!='+'&&*s!='-'&&*s!='.'&&!isdigit(*s) ) return(0);

/* ʸ */
  c=s;i=k=e=0;
  while(*c != '\0'){
    if(*c!='+'&&*c!='-'&&*c!='.'&&!isdigit(*c)&&*c!='e') return(0);
    if(isdigit(*c)) i++ ;
    if(*c=='.') k++;
    if(*c=='e') e++;
    c++;
  }
  if(i==0) return(0); /* ʤ */
  if(k>1) return(0);  /* 2İʾ */
  if(e>1) return(0);  /* e2İʾ */
  c=s;i=0;e=-1;k=0;
  while(*c != '\0'){
   if(!(i==0||e==0)&&(*c=='+'||*c=='-')) return(0); /* 椬ʰ */
   if(e>=0 && isdigit(*c) ) k++;
   if(e>=0) e++;
   if(*c=='e') e=0 ;
   i++;
   c++;
  }

  if(e==0 && k==0) return(0); /* eθ˿ʤ */
  if(k>3)  return(0); /* eθ4ʾ */

  return(1);
}
