/*=======================================================================*/
/*  q_fcombine                                                           */
/*  q_wcombine                                                           */
/*     mean, median, and stadard deviation image                         */
/*=======================================================================*/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<unistd.h>
#include<q_lib.h>
/*=======================================================================*/
void print_usage(char *);
int check_command_line(int,char **,char ***,char ***,
                         int *,char **,char **,char **,int *);
int chek_file_list(char **,int,int *,int *,int *,int *,int *);
/*=======================================================================*/
void print_usage(char *s)
{
char *ss;

  ss=strip_name(s);
  fprintf(stdout,
"NAME\n"
"       %s makes mean, median, and standard deviation image from input\n"
"       images.\n"
"SYNOPSIS\n"
"       q_fcombine file1 file2 [file3...] [msk=maskfile] [ave=avefile] \\ \n"
"       [med=medfile] [sig=sigmafile]\n"
"       q_fcombine @filelist [msk=@masklist] [ave=avefile] \\ \n"
"       [med=medfile] [sig=sigmafile]\n"
"       q_wcombine @filelist @weightlist avefile\n\n"
"DESCRIPTION\n"
"       %s makes a mean image and/or a median image and/or a standard\n"
"       deviation image from input images.\n"
"       maskfile is available not to use some pixels.\n"
"       In the first synopsis form, input image files are listed in a command\n"
"       line. In the second synopsis form, input image files are taken from\n"
"       filelist which has one file name on one line.\n"
"       Number of input files should be two or over.\n"
"       In the first synopsis form, you cannot list many input files because\n"
"       of the limit of shell.\n"
"       Output file is specified after '=' for mean image (ave), median image\n"
"       (med), and standard deviation image (sig). Do not insert space\n"
"       between '=' and file name.\n" 
"       In the first synopsis form, mask file is common for all input files\n"
"       In the second synopsis form, mask file is taken from the list.\n"
"       In this case, number of files in filelist and masklist should be\n"
"       the same.\n"
"       In the third synopsis form, this makes a mean image with weight.\n"
"       Input images are taken from files listed in filelist.\n"
"       Weight data should be an image for each input image, with the same\n"
"       dimension and size. Weight images are taken from files listed in\n"
"       weightlist. Number of files in filelist and weightlist should be\n"
"       same and should be two or over.\n"
"       Say f(k) is the value at some pixel of k-th file, and w(k) is the\n"
"       value at the same pixel, then output data on this pixel is\n"
"         m = {f(k)*w(k)}/{w(k)}.\n"
,ss,ss) ;

}
/*=======================================================================*/
int main(int argc, char ** argv)
{
char **ifile,**mask;
char *ave_file,*med_file,*sig_file;
int file_num;
int x,y,z,w,nax,iw,iz,i,ix,iy;
int mx,my,mz,mw,mnax;
FITS_DATA dat;   /* input data file */
FITS_DATA msk;   /* input mask file */
FITS_DATA ave;   /* output ave file */
FITS_DATA med;   /* output med file */
FITS_DATA sig;   /* output sig file */
FITS_DATA tmp1;  /* Working data */
FITS_DATA tmp2;  /* Working mask */
char tmpdir[128];
char tmpdat[128];
char tmpmsk[128];
int status=0;
REGION reg;
int mode;

/* Check command line */
  if(check_command_line(argc,argv,&ifile,&mask,&file_num,
                        &ave_file,&med_file,&sig_file,&mode)!=0){
    fprintf(stdout,"Error @ check_command_line\n");exit(1);}

/* Check Dimmension */
  if(chek_file_list(ifile,file_num,&x,&y,&z,&w,&nax)!=0) {
    fprintf(stdout,"fcombine error  chek_file_list of ifile\n");
    exit(-1) ;
  }
  if(mask[0]!=NULL){
    if(chek_file_list(mask,file_num,&mx,&my,&mz,&mw,&mnax)!=0) {
      fprintf(stdout,"fcombine error  chek_file_list of mask\n");
      exit(-1) ;
    }
    if(x!=mx||y!=my||z!=mz||w!=mw||nax!=mnax){
      fprintf(stdout,"fcombine error  Dimension ifile != mask  \n");
      exit(-1) ;
    }
  }
  reg.x0=1;reg.x1=x; reg.y0=1;reg.y1=y; reg.xr=REG_LIST ; reg.yr=REG_LIST;
  reg.z0=1;reg.z1=file_num;reg.zr=REG_REGION;
  reg.w0=1;reg.w1=1;reg.wr=REG_POINT;

/* Open output file and alloc memory */
  if(ave_file!=NULL){
    if(createfits(&ave,ave_file)!=0){ exit(1); } 
    ave.naxes[0] = x ; ave.naxes[1] = y ; ave.naxes[2] = z ; ave.naxes[3] = w ;
    ave.bp = DOUBLE_IMG ; ave.nax = nax ;
    if(alloc_memory_for_data(&ave)!=0) exit(1);
    if(write_head_img(&ave)!=0) exit(1);
  }
  if(med_file!=NULL){
    if(createfits(&med,med_file)!=0){ exit(1); } 
    med.naxes[0] = x ; med.naxes[1] = y ; med.naxes[2] = z ; med.naxes[3] = w ;
    med.bp = DOUBLE_IMG ; med.nax = nax ;
    if(alloc_memory_for_data(&med)!=0) exit(1);
    if(write_head_img(&med)!=0) exit(1);
  }
  if(sig_file!=NULL){
    if(createfits(&sig,sig_file)!=0){ exit(1); } 
    sig.naxes[0] = x ; sig.naxes[1] = y ; sig.naxes[2] = z ; sig.naxes[3] = w ;
    sig.bp = DOUBLE_IMG ; sig.nax = nax ;
    if(alloc_memory_for_data(&sig)!=0) exit(1);
    if(write_head_img(&sig)!=0) exit(1);
  }

/* Temporary directory */
  /* ǥ쥯ȥ˺ȥե롣
     3ܤĤäեޤȤᡢȤ */
  /* ǥ쥯ȥ̾ */
  sprintf(tmpdir,"%s/q_lib_XXXXXX",P_tmpdir);
  if(mkdtemp(tmpdir)==NULL){fprintf(stdout,"Cannot make tmp dir\n");exit(1);}
/* Data Loop */
  for(iw=0;iw<w;++iw){
  for(iz=0;iz<z;++iz){
    sprintf(tmpdat,"%s/data%d%d",tmpdir,iz,iw);
    sprintf(tmpmsk,"%s/mask%d%d",tmpdir,iz,iw);
    tmp1.naxes[0]=x;tmp1.naxes[1]=y;tmp1.naxes[2]=file_num;tmp1.naxes[3]=1;
    tmp1.bp = DOUBLE_IMG ; tmp1.nax = 3 ;
    if(alloc_memory_for_data(&tmp1)!=0) exit(1);
    if(createfits(&tmp1,tmpdat)!=0){fprintf(stdout,"Error CF\n");exit(1); }
    if(write_head_img(&tmp1)!=0){  fprintf(stdout,"Error WH\n");exit(1); }
    if(mask[0]!=NULL){
      tmp2.naxes[0]=x;tmp2.naxes[1]=y;tmp2.naxes[2]=file_num;tmp2.naxes[3]=1;
      tmp2.bp = DOUBLE_IMG ; tmp2.nax = 3 ;
      if(alloc_memory_for_data(&tmp2)!=0) exit(1);
      if(createfits(&tmp2,tmpmsk)!=0){fprintf(stdout,"Error CFm\n");exit(1); }
      if(write_head_img(&tmp2)!=0){  fprintf(stdout,"Error WHm\n");exit(1); }
    }
    /* ϥեɤǰե˽ */
    for(i=0;i<file_num;++i){
      if(openfits(&dat,ifile[i])!=0){ fprintf(stdout,"Error OF\n");exit(1);} 
      if(prep_readfits(&dat)!=0){ fprintf(stdout,"Error PR\n");exit(1);}
      if(read_2d_data(&dat,iz+1,iw+1)!=0){fprintf(stdout,"Error 2D\n");exit(1);}
      for(iy=0;iy<y;++iy){
      for(ix=0;ix<x;++ix){
        ((double *)(tmp1.data))[ix+iy*x] = d_data(&dat,ix,iy);
      }}
      fits_close_file(dat.fp,&status) ; free(dat.data);
      if(write_2d_data(&tmp1,i+1,1)!=0){fprintf(stdout,"Error WD\n");exit(1);}
      if(mask[0]!=NULL){
        if(openfits(&msk,mask[i])!=0){ fprintf(stdout,"Error OFm\n");exit(1);}
        if(prep_readfits(&msk)!=0){ fprintf(stdout,"Error PR\n");exit(1);}
        if(read_2d_data(&msk,iz+1,iw+1)!=0){fprintf(stdout,"Er 2Dm\n");exit(1);}
        for(iy=0;iy<y;++iy){
        for(ix=0;ix<x;++ix){
          ((double *)(tmp2.data))[ix+iy*x] = d_data(&msk,ix,iy);
        }}
        fits_close_file(msk.fp,&status) ; free(msk.data);
        if(write_2d_data(&tmp2,i+1,1)!=0){fprintf(stdout,"Err WDm\n");exit(1);}
      }
    }
    fits_close_file(tmp1.fp,&status) ; free(tmp1.data) ;
    if(mask[0]!=NULL){fits_close_file(tmp2.fp,&status);free(tmp2.data);}
  /* calc tmp file */
    if(openfits(&tmp1,tmpdat)!=0){ exit(1); }
    if(prep_readfits(&tmp1)!=0){ exit(1); }
    if(mask[0]!=NULL){
      if(openfits(&tmp2,tmpmsk)!=0){ exit(1); }
      if(prep_readfits(&tmp2)!=0){ exit(1); }
    }
    if(sig_file!=NULL && ave_file!=NULL){
      if(mask[0]!=NULL){
        q_stat(&tmp1,&ave,&sig,&reg,FCONV_AVE | FCONV_SIG,&tmp2,0.0 ) ;
      }else{
        q_stat(&tmp1,&ave,&sig,&reg,FCONV_AVE | FCONV_SIG,NULL,0.0 ) ;
      }
      if(write_2d_data(&ave,iz+1,iw+1)!=0) { exit(1); }
      if(write_2d_data(&sig,iz+1,iw+1)!=0) { exit(1); }
    }else if(ave_file!=NULL){
      if(mask[0]!=NULL){ /* q_wcombineϤˤʤ */
        if(mode==0)q_stat(&tmp1,&ave,NULL,&reg,FCONV_AVE|FCONV_WTE,&tmp2,0.0);
        else       q_stat(&tmp1,&ave,NULL,&reg,FCONV_AVE ,&tmp2,0.0 ) ;
      }else{
        q_stat(&tmp1,&ave,NULL,&reg,FCONV_AVE,NULL ,0.0 ) ;
      }
      if(write_2d_data(&ave,iz+1,iw+1)!=0) { exit(1); }
    }else if(sig_file!=NULL){
      if(mask[0]!=NULL){
        q_stat(&tmp1,NULL,&sig,&reg,FCONV_AVE| FCONV_SIG,&tmp2,0.0 ) ;
      }else{
        q_stat(&tmp1,NULL,&sig,&reg,FCONV_AVE| FCONV_SIG,NULL,0.0 ) ;
      }
      if(write_2d_data(&sig,iz+1,iw+1)!=0) { exit(1); }
    }
    if(med_file!=NULL){
      if(mask[0]!=NULL){
        q_stat(&tmp1,&med,NULL,&reg,FCONV_MED,&tmp2,0.0 ) ;
      }else{
        q_stat(&tmp1,&med,NULL,&reg,FCONV_MED,NULL,0.0 ) ;
      }
      if(write_2d_data(&med,iz+1,iw+1)!=0) { exit(1); }
    }
    unlink(tmpdat);
    unlink(tmpmsk);
  }}

  if(ave_file!=NULL) fits_close_file(ave.fp,&status) ;
  if(sig_file!=NULL) fits_close_file(sig.fp,&status) ;
  if(med_file!=NULL) fits_close_file(med.fp,&status) ;

  if(rmdir(tmpdir)!=0){fprintf(stdout,"Error at remove tempdir\n");}

  exit(0);
}
/*=======================================================================*/
int check_command_line(int argc,char **argv,char ***ifname,char ***mask,
    int *file_num,char ** ave_file, char ** med_file, char **sig_file,int *mode)
{
int i,j;
int fave,fsig,fmed,fmsk,nmsk;
char *ss;

  if(argc==1){/* This is normal usage query  */
     print_usage(argv[0]);exit(0);}

  ss=strip_name(argv[0]);
  fave=fsig=fmed=fmsk=nmsk=0;
  *ave_file = *med_file = *sig_file = NULL ;

  if(strncmp(ss,"q_fcombine",10)==0){  *mode=1;
  if(argc<3 ){ /* Ǥ2İʾΰɬ */
    print_usage(argv[0]); return(-1); }

  if(strncmp("@",argv[1],1)!=0) { /* file̾ľܻ */
    for(i=1;i<argc;i++)
      if(strncmp("ave=",argv[i],4)==0 || strncmp("med=",argv[i],4)==0 
        || strncmp("sig=",argv[i],4)==0 || strncmp("msk=",argv[i],4)==0 ) break;
    if(i==argc){
      fprintf(stdout,"No output file\n");return(-1);
    }
    if(i<=2) {
      fprintf(stdout,"Too few input files\n");return(-1);
    }
    j=i;
    *file_num = i-1 ;
    *ifname = (char **)malloc(sizeof(char *)*(*file_num));
    for(i=0;i<*file_num;i++) (*ifname)[i] = argv[i+1] ; 
  } else {                        /* filelistλ   */
    j=2;
    if((*file_num=read_list((argv[1]+1),ifname))<0) {
      fprintf(stdout,"Error @ read_list.\n") ; return(-1) ;
    }
  }

  for(i=j;i<argc;++i){
    if(strncmp("ave=",argv[i],4)==0 && *(argv[i]+4)!='\0' ) {
           *ave_file = (argv[i]+4) ; fave=1 ; }
    if(strncmp("med=",argv[i],4)==0 && *(argv[i]+4)!='\0' ) {
            *med_file = (argv[i]+4) ; fmed=1 ; }
    if(strncmp("sig=",argv[i],4)==0 && *(argv[i]+4)!='\0' ) {
            *sig_file = (argv[i]+4) ; fsig = 1 ; }
    if(strncmp("msk=",argv[i],4)==0 && *(argv[i]+4)!='\0' ) {
      if(*(argv[i]+4)=='@'){
        nmsk=read_list((argv[i]+5),mask) ;
        if(nmsk!=*file_num){
          fprintf(stdout,"Error @ read_list for mask \n"); return(-1) ;
        }
      }else {
        *mask = (char **)malloc(sizeof(char *)*(*file_num));
        for(i=0;i<*file_num;i++) (*mask)[i] = argv[i]+4 ;
      }
      fmsk=1;
    }
  }
  if(fave == 0 && fmed == 0 && fsig == 0){
    fprintf(stdout,"No output file\n");return(-1);
  }
  if(fmsk==0){
    *mask = (char **)malloc(sizeof(char *)*(*file_num));
    for(i=0;i<*file_num;i++) (*mask)[i] = NULL ;
  }
  }else{ *mode=0;// q_wcombine
  if(argc!=4 ){ 
    print_usage(argv[0]); return(-1); }
  if(strncmp("@",argv[1],1)!=0) { 
     fprintf(stderr,"Inputfile must be a list.\n");return(-1);
  }
  if(strncmp("@",argv[2],1)!=0) { 
     fprintf(stderr,"Inputfile must be a list.\n");return(-1);
  }
  if((*file_num=read_list((argv[1]+1),ifname))<0) {
    fprintf(stdout,"Error @ read_list.\n") ; return(-1) ;
  }
  nmsk=read_list((argv[2]+1),mask) ;
  if(nmsk!=*file_num){
    fprintf(stdout,"Error @ read_list for weight\n"); return(-1) ;
  }
  *ave_file = argv[3] ;

  }
  
  set_cmd_hist(argc,argv);

  return(0);
}
/*=======================================================================*/
/* եꥹȤμץå                                    */
int chek_file_list(char **ifile,int file_num,
                          int *x,int *y,int *z,int *w,int *nax)
{
FITS_DATA data;
int i;
int status=0;

/* read first file */
  if(openfits(&data,ifile[0])!=0){ return(1);}
/* Read header of the input files */
  if(prep_readfits(&data)!=0){ return(1); }
  *x = data.naxes[0] ; *y = data.naxes[1] ;
  *z = data.naxes[2] ; *w = data.naxes[3] ;
  *nax = data.nax;
  fits_close_file(data.fp,&status);

  for(i=1;i<file_num;++i){
    if(openfits(&data,ifile[i])!=0){ return(1);}
    if(prep_readfits(&data)!=0){ return(1); }
    if(*z!=data.naxes[2]){
      fprintf(stdout,"Z dim of %dth file is not the same as others\n",i);
      return(1);
    }
    if(*w!=data.naxes[3]){
      fprintf(stdout,"W dim of %dth file is not the same as others\n",i);
      return(1);
    }
    if(*y!=data.naxes[1]){
      fprintf(stdout,"Y dim of %dth file is not the same as others\n",i);
      return(1);
    }
    if(*x!=data.naxes[0]){
      fprintf(stdout,"X dim of %dth file is not the same as others\n",i);
      return(1);
    }
    if(*nax!=data.nax){
      fprintf(stdout,"NAXIS of %dth file is not the same as others\n",i);
      return(1);
    }

    fits_close_file(data.fp,&status);
  }
  return(0);
}
