/*=======================================================================*/
/*  q_mkimg                                                              */
/*     ǡ󤫤                                            */
/*=======================================================================*/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<q_lib.h>
/*=======================================================================*/
void print_usage(char *);
int check_command_line(int,char **,char ***,char ***,FITS_DATA *,int *);
/*=======================================================================*/
void print_usage(char *s)
{
char *ss;

  ss=strip_name(s);
  fprintf(stdout,
"NAME\n"
"       %s - Make an simple FITS image file from pixel value data.\n\n"
"SYNOPSIS\n"
"       %s   data_file       fits_file        x y [z [w]] [TYPE]\n"
"       %s @data_file_list  @fits_file_list   x y [z [w]] [TYPE]\n\n"
"DESCRIPTION\n"
"       %s makes an simple FITS image from the pixel values in data_file.\n"
"       x, y, z, w axis size should be specified by command line by integer.\n"
"       Without w specification, nax=3 image is created. Without w and x,\n"
"       nax=2 image is created.\n"
"       TYPE is one of BYTE, SHORT, LONG, LONGLONG, FLOAT, or DOUBLE.\n"
"       DOUBLE is the default.\n"
"       In data_file, one line has one pixel value and reading loop nest\n"
"       use x,y,z,w from inseide to outside, like followings.\n"
"          value at w=1,z=1,y=1,x=1\n"
"          value at w=1,z=1,y=1,x=2\n"
"             .............\n"
"          value at w=1,z=1,y=1,x=x\n"
"          value at w=1,z=1,y=2,x=1\n"
"          value at w=1,z=1,y=2,x=2\n"
"             .............\n"
"          value at w=1,z=1,y=2,x=x\n"
"             .............\n"
"          value at w=1,z=1,y=y,x=x\n"
"          value at w=1,z=2,y=1,x=1\n"
"             .............\n"
"             .............\n"
"          value at w=1,z=z,y=y,x=x\n"
"          value at w=2,z=1,y=1,x=1\n"
"             .............\n"
"          value at w=w,z=z,y=y,x=x\n\n"
"       @file_list use files written in the file_list.\n\n"
,ss,ss,ss,ss) ;
}
/*=======================================================================*/
int main(int argc, char ** argv)
{
char **ifile;
char **ofile;
FITS_DATA data;
int file_num,i,j,z,w,status=0;
FILE *fpi;
unsigned char *b;int16_t *s;int32_t *h;int64_t *l;
float *f;double *d;

/* Check command line */
  if(check_command_line(argc,argv,&ifile,&ofile,&data,&file_num)!=0){
    fprintf(stdout,"Error @ check_command_line\n");exit(1);}

/* Allocate memory for data */
  if(alloc_memory_for_data(&data)!=0) return(-1) ;
  b = (unsigned char *)data.data ; s = (int16_t *)data.data ;
  h = (int32_t *)data.data ; l = (int64_t *)data.data ;
  f = (float   *)data.data ; d = (double  *)data.data ;

/* Main loop */
  for(i=0;i<file_num;i++) {
  /* Open input file and out put files */
    if(createfits(&data,ofile[i])!=0){ exit(1); }
    if(write_head_img(&data)!=0){ exit(1); }

    if((fpi=fopen(ifile[i],"r"))==NULL) {
      fprintf(stdout,"Cannot open %s\n",ifile[i]);exit(1);}

  /* Z W loop */
    for(w=1;w<=data.naxes[3];++w){
    for(z=1;z<=data.naxes[2];++z){
      for(j=0;j<data.naxes[0]*data.naxes[1];j++) {
        if(data.bp == BYTE_IMG ){ fscanf(fpi,"%hhd",&(b[j])) ;}
        else if(data.bp == SHORT_IMG){ fscanf(fpi,"%"SCNd16"",&(s[j])) ;}
        else if(data.bp == LONG_IMG ){ fscanf(fpi,"%"SCNd32"",&(h[j])) ;}
        else if(data.bp == LONGLONG_IMG){ fscanf(fpi,"%"SCNd64"",&(l[j])) ;}
        else if(data.bp == FLOAT_IMG){ fscanf(fpi,"%f",&(f[j])) ;}
        else if(data.bp == DOUBLE_IMG){ fscanf(fpi,"%lf",&(d[j])) ;}
        else {fprintf(stdout,"Bad BP type\n");exit(1);}
      }
      if(write_2d_data(&data,z,w)!=0){
        fprintf(stdout,"Error at write_2d_data\n");exit(1);}
    }}
   /* Close */
    fits_close_file(data.fp,&status);
  }

  exit(0);
}
/*=======================================================================*/
int check_command_line(int argc,char **argv,char ***ifname,char *** ofname,
                        FITS_DATA *data,int *file_num)
{
int k,x,y,z,w;

  if(argc==1){/* This is normal usage query  */
     print_usage(argv[0]);exit(0);}

  if(argc>=5 && argc <= 8){
    if(strncmp("@",argv[1],1)!=0) { /* file̾ľܻ */
      *file_num=1 ;
      *ifname = (char **)malloc(sizeof(char *)*(*file_num));
      *ofname = (char **)malloc(sizeof(char *)*(*file_num));
      (*ifname)[0] = argv[1] ;
      (*ofname)[0] = argv[2] ;
    } else {                        /* filelistλ   */
      if((*file_num=read_list((argv[1]+1),ifname))<0) {
        fprintf(stdout,"Error @ read_list.\n") ; return(-1) ;
      }
      if((k=read_list((argv[2]+1),ofname))<0) {
        fprintf(stdout,"Error @ read_list.\n") ; return(-1) ;
      }
      if(k!=*file_num){
        fprintf(stdout,"Error @ read_list. #in!=#out\n") ; return(-1) ;
      }
    }
  } else {
    print_usage(argv[0]); return(-1);
  }
  if(strncmp(argv[argc-1],"BYTE",4)==0) { data->bp = BYTE_IMG ;}
  else if(strncmp(argv[argc-1],"SHORT",5)==0) { data->bp = SHORT_IMG ;}
  else if(strncmp(argv[argc-1],"LONGLONG",8)==0) { data->bp = LONGLONG_IMG ;}
  else if(strncmp(argv[argc-1],"LONG",4)==0) { data->bp = LONG_IMG ;}
  else if(strncmp(argv[argc-1],"FLOAT",5)==0) { data->bp = FLOAT_IMG ;}
  else if(strncmp(argv[argc-1],"DOUBLE",6)==0){ data->bp = DOUBLE_IMG ;}
  else { data->bp = DOUBLE_IMG ; argc++;}
  argc--;
  
  sscanf(argv[3],"%d",&x);
  sscanf(argv[4],"%d",&y);
  data->naxes[0] = x ; data->naxes[1] = y ;
  if(argc==5){
    data->nax = 2 ; data->naxes[2] = 1 ; data->naxes[3] = 1;
  }else if(argc==6){
    sscanf(argv[5],"%d",&z);
    data->nax = 3 ; data->naxes[2] = z ; data->naxes[3] = 1;
  }else if(argc==7){
    sscanf(argv[5],"%d",&z);
    sscanf(argv[6],"%d",&w);
    data->nax = 4 ; data->naxes[2] = z ; data->naxes[3] = w;
  }else {
    print_usage(argv[0]); return(-1);
  }

  set_cmd_hist(argc,argv);

  return(0);
}
