/*=======================================================================*/
/*  q_photo                                                              */
/*     Radial profile plot                                         */
/*=======================================================================*/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<unistd.h>
#include<math.h>
#include<signal.h>
#include<q_lib.h>
/*=======================================================================*/
void print_usage(char *);
int check_command_line(int,char **,char **,REGION *);
static void sig_handler(int);

int max_calc_time;
int photo_analysis(FITS_DATA *,double *,double *,int,int,int,int,int,
                                           double *,double *,int,char *);
void fit_photo_1(FITS_DATA *,int,int,int,int,double *,double *,double *,double *,double *);
double z_photo5(FITS_DATA *,int,int,int,int,
   double*,double,double *,double,double*,double *,double,double *,double);
double z_photo4(FITS_DATA *,int,int,int,int,
     double*,double,double *,double,double*,double,double,double *,double);
double z_photo3(FITS_DATA *,int,int,int,int,
            double*,double,double *,double,double,double,double *,double);
double z_photo2(FITS_DATA *,int,int,int,int,
                   double *,double,double,double,double,double *,double);
double z_photo1(FITS_DATA *,int,int,int,int,
                          double,double,double,double,double *,double);
double zansa_photo(FITS_DATA *,int,int,int,int,
                   double,double,double,double,double);
/*=======================================================================*/
static void sig_handler(int n)
{
  printf("It takes more than %d seconds. Aborting\n",max_calc_time);
  exit(1);
}
/*=======================================================================*/
void print_usage(char *s)
{
char *ss;

  ss=strip_name(s);
  fprintf(stdout,
"NAME\n"
"       q_photo - Radial profile plot\n\n"
"SYNOPSIS\n"
"       q_photo file_name w_region x_region y_region z_region [max_time]\n"
"       q_photo_ps file_name w_region x_region y_region z_region [max_time]\n"
"       q_photo_gif file_name w_region x_region y_region z_region [max_time]\n\n"
"DESCRIPTION\n"
"       q_photo  file_name Ǽǡ radial profile plot\n"
"       ȤᡢץåȤϤޤ\n"
"       ޤϷ̤ξɸϤ˽Ϥޤ \n\n"
"       ϤϰϤ x_region  y_region  ֿ:פȽ񤤤\n"
"       ΰꤷƲΰ̣ϳޤֿ-\n"
"       Ƚ񤤤ƥꥹȻꤷƤդޤ\n"
"       w_regionz_regionñʤֿפ񤤤ͤꤷƲ\n\n"
"       q_photo_ps  q_photo_gif Ͻϲ qplot.gif qplot.ps Ȥ\n"
"       եGIF, PSǽϤޤq_photoǤXβ̤˽Ϥޤ\n\n"
"       ׻򤢤 max_time äꤹ뤳ȤǤޤ\n\n"
          );
}
/*=======================================================================*/
int main(int argc, char ** argv)
{
char *fits_file;
FITS_DATA fits;
REGION reg;
int status=0;
int n,nt;
double *x,*y,*r,*v;
int tmpfd1,tmpfd2;
char tmpfile1[FLEN_CARD],tmpfile2[FLEN_CARD];
FILE *fp1,*fp2,*gp;
int i;
char label[FLEN_CARD*5];
char l[FLEN_CARD];
char *c,*cc;

/* Check command line */
  if(check_command_line(argc,argv,&fits_file,&reg)!=0){
    fprintf(stdout,"Error @ check_command_line\n");exit(1);}

/* Open input file */
  if(openfits(&fits,fits_file)!=0){ exit(1); }
/* Read header of the input file  */
  if(prep_readfits(&fits)!=0){ exit(1); }

/* Set timer */
  if(max_calc_time>0){
    signal(SIGALRM,sig_handler);
    alarm(max_calc_time);
  }

/* check data type */
  if(set_region(&reg,&fits)!=0) { exit(1); }

/* x dim */
  if((reg.xr!=REG_REGION)&&(reg.xr!=REG_LIST)){
    fprintf(stdout,"X region error\n");exit(-1);
  }
/* y dim */
  if((reg.yr!=REG_REGION)&&(reg.yr!=REG_LIST)){
    fprintf(stdout,"Y region error\n");exit(-1);
  }
/* z dim */
  if(reg.zr!=REG_POINT){
    fprintf(stdout,"Z region error\n");return(-1);
  }
/* w dim */
  if(reg.wr!=REG_POINT){
    fprintf(stdout,"W region error\n");return(-1);
  }
  
/* Alloc memory */
  n = (reg.x1-reg.x0+1)*(reg.y1-reg.y0+1) ;
  if((x=(double *)malloc(n*sizeof(double)))==NULL){
    fprintf(stdout,"malloc error [a]\n");return(-1);}
  if((y=(double *)malloc(n*sizeof(double)))==NULL){
    fprintf(stdout,"malloc error [b]\n");return(-1);}
  nt=100;
  if((r=(double *)malloc(nt*sizeof(double)))==NULL){
    fprintf(stdout,"malloc error [r]\n");return(-1);}
  if((v=(double *)malloc(nt*sizeof(double)))==NULL){
    fprintf(stdout,"malloc error [v]\n");return(-1);}

/* Read data */
  if(read_2d_data(&fits,reg.z0,reg.w0)!=0) return(-1) ;

/* photo analysis */
  if(photo_analysis(&fits,x,y,reg.x0-1,reg.x1-1,reg.y0-1,reg.y1-1,
                                          n,r,v,nt,label)!=0) exit(1);

/* plot */
 /* ե */
  sprintf(tmpfile1,"%s/gp_XXXXXX",P_tmpdir); 
  tmpfd1 = mkstemp(tmpfile1);
  if((fp1=fdopen(tmpfd1,"w"))==NULL){
      fprintf(stdout,"Cannot open tmp file\n");unlink(tmpfile1);exit(-1);}
  sprintf(tmpfile2,"%s/gp_XXXXXX",P_tmpdir);
  tmpfd2 = mkstemp(tmpfile2);
  if((fp2=fdopen(tmpfd2,"w"))==NULL){
      fprintf(stdout,"Cannot open tmp file\n");unlink(tmpfile2);exit(-1);}

 /* ץåȥǡ */
  for(i=0;i<n;++i) fprintf(fp1,"%e %e\n",x[i],y[i]);
  for(i=0;i<nt;++i) fprintf(fp2,"%e %e\n",r[i],v[i]);
  fclose(fp1);fclose(fp2);

/* gnuplotѥפǥץ󡢥ɥäʤ⡼ɤˤ */
  if((gp=popen("gnuplot -persist","w"))==NULL){
    fprintf(stdout,"Cannot open gnuplot\n");return(1);
  }
  if(reg.cntl & XFC_PLOT_GIF){
    fprintf(gp,"set terminal gif size 640,480\n");
    fprintf(gp,"set output \"%s\"\n",XFC_GIF_FILE);
  }else if (reg.cntl & XFC_PLOT_PS){
    fprintf(gp,"set terminal postscript eps enhanced color\n");
    fprintf(gp,"set output \"%s\"\n",XFC_PS_FILE);
  }else{
    fprintf(gp,"set terminal x11\n");
  }
/* 襳ޥɤ */
  c = label ; i=0;
  while( *c != '\0' ){
    cc = l ;
    while(*c!='\0'&&*c!='\n'){
      *cc = *c ; c++; cc++;
    }
    *cc = '\0' ;
    fprintf(gp,"set label \"%s\" at graph 0.5, graph %.2f\n",
                                         l,0.6-(double)i/20.0);
    if(*c=='\n') c++; 
    i++;
  }
  fprintf(gp,"plot \"%s\"using 1:2 title \"%s\", \"%s\" using 1:2 title \" \" with li\n",tmpfile1,fits_file,tmpfile2);

  fflush(gp); /* Don't forget to flush the buffer. */
  pclose(gp);


 /* եõ */
  unlink(tmpfile1); unlink(tmpfile2);

  printf("q_photo end\n");
/* bye */
  fits_close_file(fits.fp,&status);
  exit(0);
}
/*=======================================================================*/
/* Radial profile plot β                                    */
int photo_analysis(FITS_DATA *data,double *r,double *d, int x0, int x1,
         int y0, int y1, int n, double *rr, double *dd,int nt,char *label)
{
int x,y,nn;
double A; /* ԡ */
double B; /* Хå饦ɤ */
double s; /*  */
double cx,cy; /* 濴κɸ */
double w,hw,tc;
double rmax;
double sB,sn;
int fit;
double in4,out6,sout6,in2,in1;
int n1,n2,n4,n6;
char l[FLEN_CARD];

/* ޤ줿ΰΥǡͤ */
/* ϥǤ顢դͤʿѤ B */
  B=0.0;nn=0;
  for(x=x0;x<=x1;++x){
    y=y0;B += d_data(data,x,y);++nn;
    y=y1;B += d_data(data,x,y);++nn;
  }
  for(y=y0+1;y<y1;++y){
    x=x0;B += d_data(data,x,y);++nn;
    x=x1;B += d_data(data,x,y);++nn;
  }
  B /= (double)nn ;
/* BΥΥ */
  sB=0.0;
  for(x=x0;x<=x1;++x){
    y=y0;sB += ((d_data(data,x,y)-B)*(d_data(data,x,y)-B));
    y=y1;sB += ((d_data(data,x,y)-B)*(d_data(data,x,y)-B));
  }
  for(y=y0+1;y<y1;++y){
    x=x0;sB += ((d_data(data,x,y)-B)*(d_data(data,x,y)-B));
    x=x1;sB += ((d_data(data,x,y)-B)*(d_data(data,x,y)-B));
  }
  if(nn>1) {sB /= (double)(nn-1);sB=sqrt(sB);}
  else sB=1.0e40 ;

/* Υ򿮤ȤĴ٤ */
  tc=0.0;nn=0;
  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    tc += (d_data(data,x,y)-B);nn++;
  }}

/* Ȥǥԡõ濴ˤ褦 */
  A=0.0;
  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    if(tc>0.){ 
       if(A<d_data(data,x,y)-B) {A=d_data(data,x,y)-B;cx=x;cy=y;}
    }else{
       if(A>d_data(data,x,y)-B) {A=d_data(data,x,y)-B;cx=x;cy=y;}
    }
  }}

/* ơϳڴŪ˹ͤ 2ԥˤ褦 */
  s = 2.0 ;

/* ǤȤSN */
  sn = tc / (sB*sqrt((double)nn)) ;
  printf("SN=%f\n",sn);

/* λ SN 3 ʲʤեåȤ̵̤ */
  if(sn>3.0 || sn<-3.0){ 
    fit = 1 ; 
  }else { 
    fit = 0 ;
  }

/* ʾνͤȤäƥեåƥ󥰤򤫤 */
  if(fit){
    fit_photo_1(data,x0,x1,y0,y1,&A,&B,&cx,&cy,&s) ;
  }else {
    A = 0.0 ; /* եåȤʤʤA=0 */
  }

/* 濴ޤäΤǥդΤ󤬤Ĥ */
  nn=0;rmax=0.;
  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    w = sqrt(((double)x-cx)*((double)x-cx)+((double)y-cy)*((double)y-cy));
    r[nn] = w ;
    if(r[nn]>rmax) rmax = r[nn] ;
    w = d_data(data,x,y);
    d[nn] = w ;
    nn++;
  }}

/* ǥΥץåȤ */
  for(x=0;x<nt;++x){
    rr[x] = x/(double)nt*rmax ;
    dd[x] = (A*exp( - (double)(rr[x]*rr[x])/s/s ) + B) ; 
  }

/* ׾׻ */
/* Ⱦ */
  hw = sqrt(-log(0.5)*s*s);

/*  */
  tc=0.0;
  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    tc += (d_data(data,x,y)-B);
  }}

/* HWHMͤȤäƷ׻ʤ */
/*  HWHM4ܤȾ°ʲΰȻפäƤΥ */
  in4=0.0 ; n4 = 0 ;
  in2=0.0 ; n2 = 0 ;
  in1=0.0 ; n1 = 0 ;
/*  HWHM6ܤȾ°ʾΰ򥹥ȻפäƤΥ */
  out6 = 0.0 ; n6 = 0; sout6=0.0;

  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    w = sqrt(((double)x-cx)*((double)x-cx)+((double)y-cy)*((double)y-cy));
    if(w> 6.0*hw) { out6 += d_data(data,x,y); n6++;} 
    if(w<=4.0*hw) { in4  += (d_data(data,x,y)-B); n4++;} 
    if(w<=2.0*hw) { in2  += (d_data(data,x,y)-B); n2++;} 
    if(w<=1.0*hw) { in1  += (d_data(data,x,y)-B); n1++;} 
  }}
  out6 /= (double)n6 ;
  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    w = sqrt(((double)x-cx)*((double)x-cx)+((double)y-cy)*((double)y-cy));
    if(w> 6.0*hw){sout6 +=((d_data(data,x,y)-out6)*(d_data(data,x,y)-out6));}
  }}
  sout6 /= (double)(n6-1);
  sout6 = sqrt(sout6);

 /* 桼ؤνϤˤƤϺɸ1ϤޤΤǺɸͤ1䤷ɽ */
  printf(" A=%.0f C(%.1f %.1f) FWHM=%.1f\n",A,cx+1,cy+1,2.*hw);
  sprintf(l," A=%.0f C(%.1f %.1f) FWHM=%.1f\n",A,cx+1,cy+1,2.*hw);
  if(fit) {strcat(label,l);} /* եåȤƤʤʤ̣ʤ */

  printf("Source (r<=%.1fpix(1xHW)) = %.1f\n   Area = %dpix SN = %.0f\n",
          1.0*hw,in1,n1,in1/(sout6*sqrt((double)n1)));
  printf("Source (r<=%.1fpix(2xHW)) = %.1f\n   Area = %dpix SN = %.0f\n",
          2.0*hw,in2,n2,in2/(sout6*sqrt((double)n2)));
  printf("Source (r<=%.1fpix(4xHW)) = %.1f\n   Area = %dpix SN = %.0f\n",
          4.0*hw,in4,n4,in4/(sout6*sqrt((double)n4)));
  printf("Sky (r>%.1fpix(6xHW))  = %.2f\n    noise/pix = %.1f\n",
          6.*hw,out6,sout6);

  sprintf(l,"Source (r<=%.1fpix) = %.0f\n   Area=%dpix SN=%.0f\n",1.0*hw,in1,
    n1,in1/(sout6*sqrt((double)n1)));
  if(fit) {strcat(label,l);} /* եåȤƤʤʤ̣ʤ */
  sprintf(l,"Source (r<=%.1fpix) = %.0f\n   Area=%dpix SN=%.0f\n",2.0*hw,in2,
    n2,in2/(sout6*sqrt((double)n2)));
  if(fit) {strcat(label,l);} /* եåȤƤʤʤ̣ʤ */
  sprintf(l,"Source (r<=%.1fpix) = %.0f\n   Area=%dpix SN=%.0f\n",4.0*hw,in4,
    n4,in4/(sout6*sqrt((double)n4)));
  if(fit) {strcat(label,l);} /* եåȤƤʤʤ̣ʤ */
  sprintf(l,"Sky (r>%.1fpix)  =%.0f\n    noise/pix=%.0f\n",6.*hw,out6,sout6);
  if(fit) {strcat(label,l);} /* եåȤƤʤʤ̣ʤ */


  return(0);
}
/*=======================================================================*/
/* Radial profile plot Υեå(6)                               */
void fit_photo_1(FITS_DATA *data,int x0,int x1,int y0,int y1,
                 double *A, double *B, double *cx, double *cy, double *s)
{
double dA,dB,dc,ds;

  dA=(*A-*B)/100. ;
  dB=dA;
  dc = 1.;
  ds = 1.;
  z_photo5(data,x0,x1,y0,y1,A,dA,B,dB,cx,cy,dc,s,ds);

  dA=(*A-*B)/1000. ;
  dB=dA;
  dc = 0.1;
  ds = 0.1;
  z_photo5(data,x0,x1,y0,y1,A,dA,B,dB,cx,cy,dc,s,ds);

}
/*=======================================================================*/
/* Radial profile plot Υեå(5)                               */
double z_photo5(FITS_DATA *data,int x0,int x1,int y0,int y1,
                   double *A,double dA,
                   double *B,double dB,double *cx,double *cy,double dc,
                   double *s,double ds)
{
double z0,z1;
int f;

  f=0;
  z0 = z_photo4(data,x0,x1,y0,y1,A,dA,B,dB,cx,*cy,dc,s,ds);
  while(1){
    z1 = z_photo4(data,x0,x1,y0,y1,A,dA,B,dB,cx,*cy+dc,dc,s,ds);
    if(z1<z0) {
      *cy = *cy+dc ; z0=z1 ; f= 1;
    }else {
      break ;
    }
  }
  if(f==0){
    while(1){
      z1 = z_photo4(data,x0,x1,y0,y1,A,dA,B,dB,cx,*cy-dc,dc,s,ds);
      if(z1<z0) {
        *cy = *cy-dc ; z0=z1 ;
      }else {
        break ;
      }
    }
  }

  return(z0);
}
/*=======================================================================*/
/* Radial profile plot Υեå(4)                               */
double z_photo4(FITS_DATA *data,int x0,int x1,int y0,int y1,
                   double *A,double dA,
                   double *B,double dB,double *cx,double cy,double dc,
                   double *s,double ds)
{
double z0,z1;
int f;

  f=0;
  z0 = z_photo3(data,x0,x1,y0,y1,A,dA,B,dB,*cx,cy,s,ds);
  while(1){
    z1 = z_photo3(data,x0,x1,y0,y1,A,dA,B,dB,*cx+dc,cy,s,ds);
    if(z1<z0) {
      *cx = *cx+dc ; z0=z1 ; f= 1;
    }else {
      break ;
    }
  }
  if(f==0){
    while(1){
      z1 = z_photo3(data,x0,x1,y0,y1,A,dA,B,dB,*cx-dc,cy,s,ds);
      if(z1<z0) {
        *cx = *cx-dc ; z0=z1 ;
      }else {
        break ;
      }
    }
  }
  return(z0);
}
/*=======================================================================*/
/* Radial profile plot Υեå(3)                               */
double z_photo3(FITS_DATA *data,int x0,int x1,int y0,int y1,
                   double *A,double dA,
                   double *B,double dB, double cx, double cy,
                   double *s,double ds)
{
double z0,z1;
int f;

  f=0;
  z0 = z_photo2(data,x0,x1,y0,y1,A,dA,*B,cx,cy,s,ds);
  while(1){
    z1 = z_photo2(data,x0,x1,y0,y1,A,dA,*B+dB,cx,cy,s,ds);
    if(z1<z0) {
      *B = *B+dB ; z0=z1 ; f= 1;
    }else {
      break ;
    }
  }
  if(f==0){
    while(1){
      z1 = z_photo2(data,x0,x1,y0,y1,A,dA,*B-dB,cx,cy,s,ds);
      if(z1<z0) {
        *B = *B-dB ; z0=z1 ;
      }else {
        break ;
      }
    }
  }

  return(z0);
}
/*=======================================================================*/
/* Radial profile plot Υեå(2)                               */
double z_photo2(FITS_DATA *data,int x0,int x1,int y0,int y1,
                   double *A,double dA, double B, double cx, double cy,
                   double *s,double ds)
{
double z0,z1;
int f;

  f=0;
  z0 = z_photo1(data,x0,x1,y0,y1,*A,B,cx,cy,s,ds);
  while(1){
    z1 = z_photo1(data,x0,x1,y0,y1,*A+dA,B,cx,cy,s,ds);
    if(z1<z0) {
      *A = *A+dA ; z0=z1 ; f= 1;
    }else {
      break ;
    }
  }
  if(f==0){
    while(1){
      z1 = z_photo1(data,x0,x1,y0,y1,*A-dA,B,cx,cy,s,ds);
      if(z1<z0) {
        *A = *A-dA ; z0=z1 ;
      }else {
        break ;
      }
    }
  }

  return(z0);
}
/*=======================================================================*/
/* Radial profile plot Υեå(1)                               */
double z_photo1(FITS_DATA *data,int x0,int x1,int y0,int y1,
                   double A, double B, double cx, double cy,
                   double *s,double ds)
{
double z0,z1;
int f;

  f=0;
  z0 = zansa_photo(data,x0,x1,y0,y1,A,B,cx,cy,*s);
  while(1){
    z1 = zansa_photo(data,x0,x1,y0,y1,A,B,cx,cy,*s+ds);
    if(z1<z0) {
      *s = *s+ds ; z0=z1 ; f= 1;
    }else {
      break ;
    }
  }
  if(f==0){
    while(1){
      z1 = zansa_photo(data,x0,x1,y0,y1,A,B,cx,cy,*s-ds);
      if(z1<z0) {
        *s = *s-ds ; z0=z1 ;
      }else {
        break ;
      }
    }
  }

  return(z0);
}
/*=======================================================================*/
/* Radial profile plot Υեå(ĺ)                            */
double zansa_photo(FITS_DATA *data,int x0,int x1,int y0,int y1,
                   double A, double B, double cx, double cy, double s)
{
int x,y;
double z,m,r;

  z=0.0;
  for(y=y0;y<=y1;++y){
  for(x=x0;x<=x1;++x){
    r = ((double)x-cx)*((double)x-cx) + ((double)y-cy)*((double)y-cy) ;
    m = A*exp(-r/s/s) + B ; 
    r = (d_data(data,x,y)-m);
    z += (r*r);
  }}
  return(z);
}
/*=======================================================================*/
int check_command_line(int argc,char ** argv ,
                       char ** fits_file,REGION *reg)
{
  if(argc==1){/* This is normal usage query  */
     print_usage(argv[0]);exit(0);}

  if(argc!=6 && argc!=7){
    print_usage(argv[0]);
    return(1);
  }
/* ϥե̾ */
  *fits_file=argv[1];

/* ޤд򼨤ɸwɤ߼ */
  region_input(argv[2],&(reg->w0),&(reg->w1),&(reg->wr));
/* ơx,y,zɸɤ߼ */
  region_input(argv[3],&(reg->x0),&(reg->x1),&(reg->xr));
  region_input(argv[4],&(reg->y0),&(reg->y1),&(reg->yr));
  region_input(argv[5],&(reg->z0),&(reg->z1),&(reg->zr));

/* q_photo_gifΤȤ */
  if(strcmp(strip_name(argv[0]),"q_photo_gif")==0){
    reg->cntl = PLOT_ON | XFC_PLOT_GIF ;
/* q_photo_psΤȤ */
  }else if(strcmp(strip_name(argv[0]),"q_photo_ps")==0){
    reg->cntl = PLOT_ON | XFC_PLOT_PS ;
  }else{
    reg->cntl = PLOT_ON ;
  }

/* ׻ */
  if(argc==7){
    sscanf(argv[6],"%d",&max_calc_time);
  }else {
    max_calc_time = 0 ;
  }

  return(0);
}
